# This program takes as input the path to the mapping between DOIs and
# the filenames of the (compressed) tex sources.

# It will unpack and try to compile the sources and, then,
# infer page numbers and DOIs and fill all the latex tags.

# The final pdf files will be generated in the `output` directory.

# It must be in the same order as in the website.
# It should be robust for missing tags.

import re
import os
import sys
import time
import shutil
import logging
import json
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from glob import glob
from calendar import month_name
from subprocess import check_output

# our imports
import common as cm
from util import *
from osutil import *
from tagutil import *
from packutil import *
from latexutil import *


def main():
    # parse argumens
    args = parser().parse_args()

    # configure logging
    logging.basicConfig(format=cm.FMT, level=getattr(logging, args.log_level))
    cm.logger = logging.getLogger(os.path.basename(__file__))
    cm.logger.addHandler(logging.StreamHandler(open('last_run.log', 'w')))

    # make directory structure
    cm.build_dirs()

    return run_jobs(args)


class Job(object):
    '''Define a compilation job.'''
    def __init__(self, name, main=None, refs=True):
        submission, year, issue = parse_name(name)

        self.sub_num = submission
        self.name = '-'.join([submission, year, issue])
        self.pdf_name = 'paper%s-source.pdf' % self.name

        self.year = year
        self.issue = issue
        self.submission = submission

        self.refs = refs

        self.main = main
        if main is not None:
            self.dir = os.path.dirname(os.path.abspath(main))
            self.tex = os.path.basename(main)

    def compile(self, onecolumn=False):
        if self.main is None:
            raise cm.JobError('main file not found')

        if onecolumn:
            replace_in_file(self.main + '.tmp', add_cmd,
                            (r'\\onecolumn',), output=self.main)

        # compile tex sources
        with cd(self.dir):
            try:
                stdout = bytes_to_str(compile_source(self.tex[:-4], self.refs))
                cm.logger.debug(stdout.split("\n")[0])
            except SystemError as exc:
                cm.logger.error("compilation failed, output:")
                cm.logger.error(str(exc))
                raise cm.JobError("Submission %s could not be compiled" %
                                  self.submission)

        # get number of pages
        if not onecolumn:
            self.pages = get_num_pages(stdout)

            # make sure paper doesn't exceed page limit
            if self.pages > cm.PAGE_LIMIT:
                cm.logger.warning("Submission %s exceeds page limit (%s > %s)" %
                                (self.submission, self.pages, cm.PAGE_LIMIT))

        # move to compiled
        shutil.move(self.pdf, self.final)

    def get(self, extension=''):
        fname = self.tex.replace('.tex', extension)
        return os.path.join(self.dir, fname)

    @property
    def key(self):
        return job_key(self.submission, self.year, self.issue)

    @property
    def pub_fname(self):
        return self.doi.split('/')[1] + '.pdf'
    
    @property
    def pub_num(self):
        return self.doi.split('/')[1].split('-')[-1]

    @property
    def pdf(self):
        return self.get('.pdf')

    @property
    def log(self):
        return self.get('.log')

    @property
    def final(self):
        pdf = 'paper%s-source.pdf' % self.key
        return os.path.join(cm.PDFS_DIR, pdf)

    def __str__(self):
        return self.name


def create_job(packed, refs=True):
    '''Return a compilation job from a source package.'''
    uname, _ = os.path.splitext(os.path.basename(packed))
    upath = os.path.join(os.path.dirname(packed), uname)

    # unpack source
    if os.path.isdir(upath):
        shutil.rmtree(upath)
    os.mkdir(upath)
    cm.logger.info("Unpacking %s to %s" % (packed, upath))
    upath = unpack(packed, upath)

    # find main path
    cm.logger.debug('Looking for main in %s' % upath)
    main = find_main(upath)
    if main is None:
        cm.logger.error("Could not find a main tex file.")
        return -1
    cm.logger.debug('Found main: %s' % main)

    # add to jobs
    return Job(os.path.basename(packed), main, refs)


def collect(path, issue):
    '''Collect all jobs to be run.
    Unpack each source and find the main file. For each source,
    instantiate a job with its submission attributes that points
    to its main file.

    Sort by name.
    '''
    jobs = []
    paper_path = os.path.join(path, 'paper*-*-%s-source.*')
    for fpath in sorted(glob(paper_path % issue)):
        cm.logger.debug("Found %s" % fpath)
        # only process files with recognized formats
        _, extension = os.path.splitext(fpath)
        if extension not in COMPRESSION_FORMATS:
            continue
        if extension == '.tar' and len(glob('%s.*' % fpath)) > 0:
            shutil.rmtree(fpath)
            continue
        job = create_job(fpath)
        jobs.append(job)
    return jobs


def prepend_editors(editors_path, jobs):
    if os.path.isfile(editors_path):
        editors_job = create_job(editors_path, False)
        jobs.insert(0, editors_job)
    else:
        cm.logger.warning("editor's intro not found!")

def generate_meta_data(jobs):
    import yaml

    year = int(jobs[0].year)
    issue = int(jobs[0].issue)
    start_page = 1

    cm.logger.info("Generating meta data as a YAML file...")
    files = []
    paper_num_mapping = {}
    for i, job in enumerate(jobs):
        cm.logger.info("\tIncluding submission %s, \n\tTitle: %s\n\tAuthors: %s\n\tAbstract: %s"
                       % (job, job.title, ", ".join([auth_info[0] for auth_info in job.authors]), job.abstract))

        paper_csd = {
            "file": 'files/papers/issue{}/{}'.format(issue, job.pub_fname),
            "volume": year,
            "issue": issue,
            "pagemin": start_page,
            "pagemax": start_page + job.pages - 1,
            "title": job.title,
            "authors": job.authors,
            "abstract": job.abstract,
            "keywords": job.keywords,
            "doi": job.doi
        }

        paper_num_mapping[job.sub_num] = job.pub_num

        start_page += job.pages
        files.append(paper_csd)

    with open(os.path.join(cm.PDFS_DIR, "{}.{}.yaml".format(year, issue)), 'w') as f:
        yaml.dump(files, f, sort_keys=False, default_flow_style=False, allow_unicode=True)
    # with open(os.path.join(cm.PDFS_DIR, "idmap.{}.{}.yaml".format(year, issue)), 'w') as f:
    #     yaml.dump(paper_num_mapping, f, sort_keys=False, default_flow_style=False)

def generate_cds(jobs):
    from docx import Document
    from docx.enum.text import WD_ALIGN_PARAGRAPH

    document = Document()

    year = jobs[0].year
    issue = jobs[0].issue
    cover_date_month = cm.COVER_DATE_MONTH[year][int(issue) - 1]

    journ_csd = {'journal_code': cm.POPETS_ID}
    issue_csd = {
                 'volume': year,
                 'issue_num': issue,
                 'cover_date': '%s %s' % (month_name[cover_date_month], year)
                }
    csd_heading = document.add_heading('CSD', 2)
    csd_heading.alignment = WD_ALIGN_PARAGRAPH.CENTER
    document.add_paragraph(cm.JOURN_CSD.format(**journ_csd))

    issue_heading = document.add_heading('Issue Information', 2)
    issue_heading.alignment = WD_ALIGN_PARAGRAPH.CENTER
    document.add_paragraph(cm.ISSUE_CSD.format(**issue_csd))

    cm.logger.info("Generating CDS...")
    for i, job in enumerate(jobs):
        cm.logger.info("\tIncluding submission %s, titled: %s\n\tAuthors: %s"
                       % (job, job.title, job.authors))
        paper_csd = {
                    'doi': job.doi,
                    'title': job.title,
                    'authors': job.authors,
                    'abstract': job.abstract,
                    'filename': '%s' % job.pub_fname,
                    # bug in publication manual (reported):
                    #'filename': '%s-%s' % (str(i + 1).zfill(2), job.pub_fname),
                    }
        document.add_paragraph(cm.PAPER_CSD.format(**paper_csd))

    document.save(os.path.join(cm.PDFS_DIR, cm.CDS_FILENAME))


def compile_jobs(jobs, dry=False, onecolumn=False):
    errored = []
    # compile papers
    for job in jobs:
        cm.logger.info('Compiling %s' % job)
        shutil.copy(job.main, job.main + '.tmp')
        try:
            job.compile(onecolumn=onecolumn)
        except cm.JobError as exc:
            cm.logger.error('%s failed:\n\t %s' % (job, str(exc)))
            errored.append(job)
        else:
            if dry: shutil.move(job.main + '.tmp', job.main)
    success = [j for j in jobs if j not in errored]
    return errored, success

def get_info_from_json(jobs):
    errored = []
    for job in jobs:
        try:
            with open(os.path.join(job.dir, "metadata.json"), 'r', encoding='utf-8') as fi:
                metadata = json.load(fi)
            job.title = metadata["title"]
            job.authors = [[parse_author_name(a["first_name"], a["last_name"]), a["institution"]] for a in metadata["authors"]]
            job.abstract = metadata["abstract"]
            job.keywords = ', '.join(metadata["keywords"])
        except FileNotFoundError:
            cm.logger.error("Missing metadata.json! Please make sure that metadata.json is placed inside your submission and is in the same folder as main.tex.")
            errored.append(job)
        except json.decoder.JSONDecodeError:
            cm.logger.error("Malformatted metadata.json! Please check if your metadata.json is formatted correctly.")
            errored.append(job)
    success = [j for j in jobs if j not in errored]
    return errored, success

def run_jobs(args):
    cm.logger.info('Compiling sources for issue %s' % args.issue)

    # collect all jobs in path
    jobs = collect(args.sources, args.issue)
    if len(jobs) == 0:
        cm.logger.warning('no jobs found in %s!' % args.sources)
        return 0

    # prepend editor's intro
    editors_fname = 'paper00-%s-%s-editorsintro.zip' % (jobs[0].year, args.issue)
    editors_path = os.path.join(args.sources, editors_fname)
    prepend_editors(editors_path, jobs)

    # assign DOIs
    num_papers = sum(cm.NUM_PAPERS[jobs[0].year][:args.issue - 1])
    for i, job in enumerate(jobs):
        job.doi = cm.PETS_DOI[job.year] + str(num_papers + i + 1).zfill(4)

    # fill tags and compile jobs
    fill_tags(jobs)
    errored1, success = compile_jobs(jobs)

    # get pdf info from pdfs
    errored2, success = compile_jobs(success, dry=True, onecolumn=True)
    errored3, success = get_info_from_json(success)

    # fill page numbers and recompile
    set_pagenums(success)
    for job in success:
        remove_aux_files(job.main)
    errored4, success = compile_jobs(success)

    # accumulate errors
    errored = errored1 + errored2 + errored3 + errored4

    # check embeded fonts
    for job in success:
        try:
            check_and_embed_font(job, args.try_embed)
        except cm.FontError as e:
            cm.logger.error(e)
            success.remove(job)
            errored.append(job)


    # summary message
    if len(errored) > 0:
        cm.logger.info('%s/%s jobs failed' % (len(errored), len(jobs)))
        return 1
    cm.logger.info('All jobs in issue %s completed successfully!' % args.issue)

    # generate content document specification (CDS)
    # generate_cds(jobs)
    generate_meta_data(jobs)

    # concatenate all pdfs in issue
    # name_pdf = 'issue-{0}-{1}-source.pdf'
    # output_pdf = os.path.join(cm.PDFS_DIR, name_pdf)
    # success_pdf = [os.path.join(cm.PDFS_DIR, job.pdf_name) for job in success]
    # os.system(('pdfunite %s %s' % (' '.join(success_pdf), output_pdf)).format(jobs[0].year, args.issue))

    # pack for publication
    cm.logger.info("Before packing for publication:\n \
                    \t\t\t- check page numbers, widows and orphans in the issue pdf\n\
                    \t\t\t- check fonts look fine after (after embedding there could be issues)\n\
                    \t\t\t- run a spell-checker on thr CDS document.\n\n Press [y] to continue...")
    while True:
        choice = input().lower()
        if choice in ['y', 'Y']:
            break
        else:
            sys.stdout.write("Please, press 'y' to continue.\n")

    return 0


def check_and_embed_font(job, embed=True):
    try:
        check_fonts(job)
    except cm.FontError as e:
        cm.logger.error(e)
        if embed:
            embed_fonts(job)
    else:
        cm.logger.info("All fonts embedded and of OK types in %s" % job)


def check_fonts(job):
    ok_fonts = ['Type 1', 'TrueType', 'CID TrueType', 'Type 1C']
    job_pdf = os.path.join(cm.PDFS_DIR, job.pdf_name)
    output = bytes_to_str(check_output(["pdffonts", job_pdf]))
    header = output.splitlines()[1]
    indices = [m.start() for m in re.finditer(' ', header)]
    for line in output.splitlines()[2:]:
        font_name = line[0:indices[0]].strip()
        type_str = line[indices[0]:indices[1]].strip()
        is_embeded = line[indices[2]:indices[3]].strip()
        is_subset = line[indices[3]:indices[4]].strip()
        if is_embeded == "no":
            raise cm.FontError("PDF %s doesn't have font %s embedded"
                               % (job, font_name))
        if not type_str in ok_fonts:
            cm.logger.error("Font %s in %s is not of any type %s, but \'%s\'"
                               % (job, font_name, ok_fonts, type_str))
        if is_subset == 'no':
            cm.logger.error("Font %s in %s is not \'subset\'" % (job, font_name))


def embed_fonts(job):
    job_pdf = os.path.join(cm.PDFS_DIR, job.pdf_name)
    cm.logger.info('Embedding fonts for job: %s' % job)
    embed_pdf = '%s_embed' % job_pdf
    os.system("sh %s %s" % (cm.EMBED_CMD, job_pdf))
    
    if os.path.isfile(embed_pdf):
        shutil.move(embed_pdf, job_pdf)
    else:
        raise cm.FontError("Embedding of fonts failed: %s" % job_pdf)


def pack_for_publication(jobs):
    year = jobs[0].year
    issue = jobs[0].issue
    pub_dir = cm.PUB_DIR.format(**{'journal_code': cm.POPETS_ID,
                                    'volume': year,
                                    'issue': issue})
    make_dir(pub_dir)
    for job in jobs:
        pdf_path = os.path.join(cm.PDFS_DIR, job.pdf_name)
        pub_path = os.path.join(pub_dir, job.pub_fname)
        shutil.copy(pdf_path, pub_path)
    CDS_FILENAME = "{}.{}.yaml".format(year, issue)
    IDMAP_FILENAME = "idmap.{}.{}.yaml".format(year, issue)
    # shutil.copy(os.path.join(cm.PDFS_DIR, cm.CDS_FILENAME),
                # os.path.join(pub_dir, cm.CDS_FILENAME))
    shutil.copy(os.path.join(cm.PDFS_DIR, CDS_FILENAME),
                os.path.join(pub_dir, CDS_FILENAME))
    shutil.copy(os.path.join(cm.PDFS_DIR, IDMAP_FILENAME),
                os.path.join(pub_dir, IDMAP_FILENAME))
    zipdir(pub_dir)
    return pub_dir


def parser():
    fmt_class = ArgumentDefaultsHelpFormatter
    parser = ArgumentParser(description='Compile tex sources for PoPETS.',
                            formatter_class=fmt_class)

    parser.add_argument('sources',
                        type=str,
                        help='path to directory where packed tex sources are.')

    parser.add_argument('--issue', '-i',
                        required=True,
                        type=int,
                        choices=range(1, cm.NUM_ISSUES + 1),
                        help='issue that will try to compile.')

    parser.add_argument('--try-embed', '-e',
                        default=False,
                        action='store_true',
                        help='whether to try to embed fonts in pdf.')


    parser.add_argument('--log-level', '-l',
                        type=str,
                        choices=[logging.getLevelName(logging.DEBUG),
                                 logging.getLevelName(logging.INFO)],
                        default=logging.getLevelName(logging.INFO),
                        help='parsing level.')
    return parser


if __name__ == '__main__':
    try:
        sys.exit(sys.exit(main()))
    except KeyboardInterrupt:
        sys.exit(0)
